@@ -30,10 +30,10 @@
 #include <stddef.h>
 #include <stdint.h>
 #include <string.h>
+#include <errno.h>
 #include <limits.h>
 
 #include <sys/types.h>
-#include <sys/mman.h>
 #include <unistd.h>
 #include <dlfcn.h>
 #include <xf86drm.h>
@@ -44,6 +44,7 @@
 #include "gbm_driint.h"
 
 #include "gbmint.h"
+#include "loader.h"
 
 /* For importing wl_buffer */
 #if HAVE_WAYLAND_PLATFORM
@@ -104,20 +105,133 @@ dri_get_buffers_with_format(__DRIdrawable * driDrawable,
                                    count, out_count, surf->dri_private);
 }
 
+static int
+image_get_buffers(__DRIdrawable *driDrawable,
+                  unsigned int format,
+                  uint32_t *stamp,
+                  void *loaderPrivate,
+                  uint32_t buffer_mask,
+                  struct __DRIimageList *buffers)
+{
+   struct gbm_dri_surface *surf = loaderPrivate;
+   struct gbm_dri_device *dri = gbm_dri_device(surf->base.gbm);
+
+   if (dri->image_get_buffers == NULL)
+      return 0;
+
+   return dri->image_get_buffers(driDrawable, format, stamp,
+                                 surf->dri_private, buffer_mask, buffers);
+}
+
+static void
+swrast_get_drawable_info(__DRIdrawable *driDrawable,
+                         int           *x,
+                         int           *y,
+                         int           *width,
+                         int           *height,
+                         void          *loaderPrivate)
+{
+   struct gbm_dri_surface *surf = loaderPrivate;
+
+   *x = 0;
+   *y = 0;
+   *width = surf->base.width;
+   *height = surf->base.height;
+}
+
+static void
+swrast_put_image2(__DRIdrawable *driDrawable,
+                  int            op,
+                  int            x,
+                  int            y,
+                  int            width,
+                  int            height,
+                  int            stride,
+                  char          *data,
+                  void          *loaderPrivate)
+{
+   struct gbm_dri_surface *surf = loaderPrivate;
+   struct gbm_dri_device *dri = gbm_dri_device(surf->base.gbm);
+
+   dri->swrast_put_image2(driDrawable,
+                          op, x, y,
+                          width, height, stride,
+                          data, surf->dri_private);
+}
+
+static void
+swrast_put_image(__DRIdrawable *driDrawable,
+                 int            op,
+                 int            x,
+                 int            y,
+                 int            width,
+                 int            height,
+                 char          *data,
+                 void          *loaderPrivate)
+{
+   return swrast_put_image2(driDrawable, op, x, y, width, height,
+                            width * 4, data, loaderPrivate);
+}
+
+static void
+swrast_get_image(__DRIdrawable *driDrawable,
+                 int            x,
+                 int            y,
+                 int            width,
+                 int            height,
+                 char          *data,
+                 void          *loaderPrivate)
+{
+   struct gbm_dri_surface *surf = loaderPrivate;
+   struct gbm_dri_device *dri = gbm_dri_device(surf->base.gbm);
+
+   dri->swrast_get_image(driDrawable,
+                         x, y,
+                         width, height,
+                         data, surf->dri_private);
+}
+
 static const __DRIuseInvalidateExtension use_invalidate = {
-   { __DRI_USE_INVALIDATE, 1 }
+   .base = { __DRI_USE_INVALIDATE, 1 }
 };
 
 static const __DRIimageLookupExtension image_lookup_extension = {
-   { __DRI_IMAGE_LOOKUP, 1 },
-   dri_lookup_egl_image
+   .base = { __DRI_IMAGE_LOOKUP, 1 },
+
+   .lookupEGLImage          = dri_lookup_egl_image
 };
 
-const __DRIdri2LoaderExtension dri2_loader_extension = {
-   { __DRI_DRI2_LOADER, 3 },
-   dri_get_buffers,
-   dri_flush_front_buffer,
-   dri_get_buffers_with_format,
+static const __DRIdri2LoaderExtension dri2_loader_extension = {
+   .base = { __DRI_DRI2_LOADER, 3 },
+
+   .getBuffers              = dri_get_buffers,
+   .flushFrontBuffer        = dri_flush_front_buffer,
+   .getBuffersWithFormat    = dri_get_buffers_with_format,
+};
+
+static const __DRIimageLoaderExtension image_loader_extension = {
+   .base = { __DRI_IMAGE_LOADER, 1 },
+
+   .getBuffers          = image_get_buffers,
+   .flushFrontBuffer    = dri_flush_front_buffer,
+};
+
+static const __DRIswrastLoaderExtension swrast_loader_extension = {
+   .base = { __DRI_SWRAST_LOADER, 2 },
+
+   .getDrawableInfo = swrast_get_drawable_info,
+   .putImage        = swrast_put_image,
+   .getImage        = swrast_get_image,
+   .putImage2       = swrast_put_image2
+};
+
+static const __DRIextension *gbm_dri_screen_extensions[] = {
+   &image_lookup_extension.base,
+   &use_invalidate.base,
+   &dri2_loader_extension.base,
+   &image_loader_extension.base,
+   &swrast_loader_extension.base,
+   NULL,
 };
 
 struct dri_extension_match {
@@ -138,6 +252,12 @@ static struct dri_extension_match gbm_dri_device_extensions[] = {
    { NULL, 0, 0 }
 };
 
+static struct dri_extension_match gbm_swrast_device_extensions[] = {
+   { __DRI_CORE, 1, offsetof(struct gbm_dri_device, core), },
+   { __DRI_SWRAST, 1, offsetof(struct gbm_dri_device, swrast) },
+   { NULL, 0, 0 }
+};
+
 static int
 dri_bind_extensions(struct gbm_dri_device *dri,
                     struct dri_extension_match *matches,
@@ -166,16 +286,27 @@ dri_bind_extensions(struct gbm_dri_device *dri,
    return ret;
 }
 
-static int
-dri_load_driver(struct gbm_dri_device *dri)
+static const __DRIextension **
+dri_open_driver(struct gbm_dri_device *dri)
 {
-   const __DRIextension **extensions;
+   const __DRIextension **extensions = NULL;
    char path[PATH_MAX], *search_paths, *p, *next, *end;
+   char *get_extensions_name;
 
    search_paths = NULL;
+   /* don't allow setuid apps to use LIBGL_DRIVERS_PATH or GBM_DRIVERS_PATH */
    if (geteuid() == getuid()) {
-      /* don't allow setuid apps to use GBM_DRIVERS_PATH */
+      /* Read GBM_DRIVERS_PATH first for compatibility, but LIBGL_DRIVERS_PATH
+       * is recommended over GBM_DRIVERS_PATH.
+       */
       search_paths = getenv("GBM_DRIVERS_PATH");
+
+      /* Read LIBGL_DRIVERS_PATH if GBM_DRIVERS_PATH was not set.
+       * LIBGL_DRIVERS_PATH is recommended over GBM_DRIVERS_PATH.
+       */
+      if (search_paths == NULL) {
+         search_paths = getenv("LIBGL_DRIVERS_PATH");
+      }
    }
    if (search_paths == NULL)
       search_paths = DEFAULT_DRIVER_DIR;
@@ -198,24 +329,48 @@ dri_load_driver(struct gbm_dri_device *dri)
          snprintf(path, sizeof path,
                   "%.*s/%s_dri.so", len, p, dri->base.driver_name);
          dri->driver = dlopen(path, RTLD_NOW | RTLD_GLOBAL);
-         if (dri->driver == NULL)
-            fprintf(stderr, "failed to open %s: %s\n", path, dlerror());
       }
+      /* not need continue to loop all paths once the driver is found */
+      if (dri->driver != NULL)
+         break;
    }
 
    if (dri->driver == NULL) {
-      fprintf(stderr, "gbm: failed to open any driver (search paths %s)",
+      fprintf(stderr, "gbm: failed to open any driver (search paths %s)\n",
               search_paths);
-      return -1;
+      fprintf(stderr, "gbm: Last dlopen error: %s\n", dlerror());
+      return NULL;
    }
 
-   extensions = dlsym(dri->driver, __DRI_DRIVER_EXTENSIONS);
+   if (asprintf(&get_extensions_name, "%s_%s",
+                __DRI_DRIVER_GET_EXTENSIONS, dri->base.driver_name) != -1) {
+      const __DRIextension **(*get_extensions)(void);
+
+      get_extensions = dlsym(dri->driver, get_extensions_name);
+      free(get_extensions_name);
+
+      if (get_extensions)
+         extensions = get_extensions();
+   }
+
+   if (!extensions)
+      extensions = dlsym(dri->driver, __DRI_DRIVER_EXTENSIONS);
    if (extensions == NULL) {
       fprintf(stderr, "gbm: driver exports no extensions (%s)", dlerror());
       dlclose(dri->driver);
-      return -1;
    }
 
+   return extensions;
+}
+
+static int
+dri_load_driver(struct gbm_dri_device *dri)
+{
+   const __DRIextension **extensions;
+
+   extensions = dri_open_driver(dri);
+   if (!extensions)
+      return -1;
 
    if (dri_bind_extensions(dri, gbm_dri_device_extensions, extensions) < 0) {
       dlclose(dri->driver);
@@ -223,16 +378,39 @@ dri_load_driver(struct gbm_dri_device *dri)
       return -1;
    }
 
+   dri->driver_extensions = extensions;
+
    return 0;
 }
 
 static int
-dri_screen_create(struct gbm_dri_device *dri)
+dri_load_driver_swrast(struct gbm_dri_device *dri)
+{
+   const __DRIextension **extensions;
+
+   extensions = dri_open_driver(dri);
+   if (!extensions)
+      return -1;
+
+   if (dri_bind_extensions(dri, gbm_swrast_device_extensions, extensions) < 0) {
+      dlclose(dri->driver);
+      fprintf(stderr, "failed to bind extensions\n");
+      return -1;
+   }
+
+   dri->driver_extensions = extensions;
+
+   return 0;
+}
+
+static int
+dri_screen_create_dri2(struct gbm_dri_device *dri,
+                       const char *driver_name)
 {
    const __DRIextension **extensions;
    int ret = 0;
 
-   dri->base.driver_name = dri_fd_get_driver_name(dri->base.base.fd);
+   dri->base.driver_name = driver_name;
    if (dri->base.driver_name == NULL)
       return -1;
 
@@ -242,17 +420,21 @@ dri_screen_create(struct gbm_dri_device *dri)
       return ret;
    };
 
-   dri->extensions[0] = &image_lookup_extension.base;
-   dri->extensions[1] = &use_invalidate.base;
-   dri->extensions[2] = &dri2_loader_extension.base;
-   dri->extensions[3] = NULL;
+   dri->extensions = gbm_dri_screen_extensions;
 
    if (dri->dri2 == NULL)
       return -1;
 
-   dri->screen = dri->dri2->createNewScreen(0, dri->base.base.fd,
-                                            dri->extensions,
-                                            &dri->driver_configs, dri);
+   if (dri->dri2->base.version >= 4) {
+      dri->screen = dri->dri2->createNewScreen2(0, dri->base.base.fd,
+                                                dri->extensions,
+                                                dri->driver_extensions,
+                                                &dri->driver_configs, dri);
+   } else {
+      dri->screen = dri->dri2->createNewScreen(0, dri->base.base.fd,
+                                               dri->extensions,
+                                               &dri->driver_configs, dri);
+   }
    if (dri->screen == NULL)
       return -1;
 
@@ -273,6 +455,72 @@ free_screen:
    return ret;
 }
 
+static int
+dri_screen_create_swrast(struct gbm_dri_device *dri)
+{
+   int ret;
+
+   dri->base.driver_name = strdup("swrast");
+   if (dri->base.driver_name == NULL)
+      return -1;
+
+   ret = dri_load_driver_swrast(dri);
+   if (ret) {
+      fprintf(stderr, "failed to load swrast driver\n");
+      return ret;
+   }
+
+   dri->extensions = gbm_dri_screen_extensions;
+
+   if (dri->swrast == NULL)
+      return -1;
+
+   if (dri->swrast->base.version >= 4) {
+      dri->screen = dri->swrast->createNewScreen2(0, dri->extensions,
+                                                  dri->driver_extensions,
+                                                  &dri->driver_configs, dri);
+   } else {
+      dri->screen = dri->swrast->createNewScreen(0, dri->extensions,
+                                                 &dri->driver_configs, dri);
+   }
+   if (dri->screen == NULL)
+      return -1;
+
+   dri->lookup_image = NULL;
+   dri->lookup_user_data = NULL;
+
+   return 0;
+}
+
+static int
+dri_screen_create(struct gbm_dri_device *dri)
+{
+   const char *driver_name;
+
+   driver_name = loader_get_driver_for_fd(dri->base.base.fd, 0);
+   if (!driver_name)
+      return -1;
+
+   return dri_screen_create_dri2(dri, driver_name);
+}
+
+static int
+dri_screen_create_sw(struct gbm_dri_device *dri)
+{
+   const char *driver_name;
+   int ret;
+
+   driver_name = strdup("kms_swrast");
+   if (!driver_name)
+      return -errno;
+
+   ret = dri_screen_create_dri2(dri, driver_name);
+   if (ret == 0)
+      return ret;
+
+   return dri_screen_create_swrast(dri);
+}
+
 static int
 gbm_dri_is_format_supported(struct gbm_device *gbm,
                             uint32_t format,
@@ -291,7 +539,7 @@ gbm_dri_is_format_supported(struct gbm_device *gbm,
       return 0;
    }
 
-   if (usage & GBM_BO_USE_CURSOR_64X64 &&
+   if (usage & GBM_BO_USE_CURSOR &&
        usage & GBM_BO_USE_RENDERING)
       return 0;
 
@@ -303,14 +551,31 @@ gbm_dri_bo_write(struct gbm_bo *_bo, const void *buf, size_t count)
 {
    struct gbm_dri_bo *bo = gbm_dri_bo(_bo);
 
-   if (bo->image != NULL)
+   if (bo->image != NULL) {
+      errno = EINVAL;
       return -1;
+   }
 
    memcpy(bo->map, buf, count);
 
    return 0;
 }
 
+static int
+gbm_dri_bo_get_fd(struct gbm_bo *_bo)
+{
+   struct gbm_dri_device *dri = gbm_dri_device(_bo->gbm);
+   struct gbm_dri_bo *bo = gbm_dri_bo(_bo);
+   int fd;
+
+   if (bo->image == NULL)
+      return -1;
+
+   dri->image->queryImage(bo->image, __DRI_IMAGE_ATTRIB_FD, &fd);
+
+   return fd;
+}
+
 static void
 gbm_dri_bo_destroy(struct gbm_bo *_bo)
 {
@@ -321,7 +586,7 @@ gbm_dri_bo_destroy(struct gbm_bo *_bo)
    if (bo->image != NULL) {
       dri->image->destroyImage(bo->image);
    } else {
-      munmap(bo->map, bo->size);
+      gbm_dri_bo_unmap(bo);
       memset(&arg, 0, sizeof(arg));
       arg.handle = bo->handle;
       drmIoctl(dri->base.base.fd, DRM_IOCTL_MODE_DESTROY_DUMB, &arg);
@@ -367,22 +632,29 @@ gbm_dri_bo_import(struct gbm_device *gbm,
    int gbm_format;
 
    /* Required for query image WIDTH & HEIGHT */
-   if (dri->image->base.version < 4)
+   if (dri->image == NULL || dri->image->base.version < 4) {
+      errno = ENOSYS;
       return NULL;
+   }
 
    switch (type) {
 #if HAVE_WAYLAND_PLATFORM
    case GBM_BO_IMPORT_WL_BUFFER:
    {
-      struct wl_drm_buffer *wb = (struct wl_drm_buffer *) buffer;
+      struct wl_drm_buffer *wb;
 
-      if (dri->wl_drm == NULL)
+      if (!dri->wl_drm) {
+         errno = EINVAL;
          return NULL;
+      }
 
-      if (!wayland_buffer_is_drm(dri->wl_drm, buffer))
+      wb = wayland_drm_buffer_get(dri->wl_drm, (struct wl_resource *) buffer);
+      if (!wb) {
+         errno = EINVAL;
          return NULL;
+      }
 
-      image = wb->driver_buffer;
+      image = dri->image->dupImage(wb->driver_buffer, NULL);
 
       switch (wb->format) {
       case WL_DRM_FORMAT_XRGB8888:
@@ -391,6 +663,9 @@ gbm_dri_bo_import(struct gbm_device *gbm,
       case WL_DRM_FORMAT_ARGB8888:
          gbm_format = GBM_FORMAT_ARGB8888;
          break;
+      case WL_DRM_FORMAT_RGB565:
+         gbm_format = GBM_FORMAT_RGB565;
+         break;
       case WL_DRM_FORMAT_YUYV:
          gbm_format = GBM_FORMAT_YUYV;
          break;
@@ -404,18 +679,40 @@ gbm_dri_bo_import(struct gbm_device *gbm,
    case GBM_BO_IMPORT_EGL_IMAGE:
    {
       int dri_format;
-      if (dri->lookup_image == NULL)
+      if (dri->lookup_image == NULL) {
+         errno = EINVAL;
          return NULL;
+      }
 
       image = dri->lookup_image(dri->screen, buffer, dri->lookup_user_data);
+      image = dri->image->dupImage(image, NULL);
       dri->image->queryImage(image, __DRI_IMAGE_ATTRIB_FORMAT, &dri_format);
       gbm_format = gbm_dri_to_gbm_format(dri_format);
-      if (gbm_format == 0)
+      if (gbm_format == 0) {
+         errno = EINVAL;
          return NULL;
+      }
+      break;
+   }
+
+   case GBM_BO_IMPORT_FD:
+   {
+      struct gbm_import_fd_data *fd_data = buffer;
+      int stride = fd_data->stride, offset = 0;
+
+      image = dri->image->createImageFromFds(dri->screen,
+                                             fd_data->width,
+                                             fd_data->height,
+                                             fd_data->format,
+                                             &fd_data->fd, 1,
+                                             &stride, &offset,
+                                             NULL);
+      gbm_format = fd_data->format;
       break;
    }
 
    default:
+      errno = ENOSYS;
       return NULL;
    }
 
@@ -424,14 +721,15 @@ gbm_dri_bo_import(struct gbm_device *gbm,
    if (bo == NULL)
       return NULL;
 
-   bo->image = dri->image->dupImage(image, NULL);
+   bo->image = image;
 
    if (usage & GBM_BO_USE_SCANOUT)
       dri_use |= __DRI_IMAGE_USE_SCANOUT;
-   if (usage & GBM_BO_USE_CURSOR_64X64)
+   if (usage & GBM_BO_USE_CURSOR)
       dri_use |= __DRI_IMAGE_USE_CURSOR;
    if (dri->image->base.version >= 2 &&
        !dri->image->validateUsage(bo->image, dri_use)) {
+      errno = EINVAL;
       free(bo);
       return NULL;
    }
@@ -458,15 +756,19 @@ create_dumb(struct gbm_device *gbm,
 {
    struct gbm_dri_device *dri = gbm_dri_device(gbm);
    struct drm_mode_create_dumb create_arg;
-   struct drm_mode_map_dumb map_arg;
    struct gbm_dri_bo *bo;
    struct drm_mode_destroy_dumb destroy_arg;
    int ret;
+   int is_cursor, is_scanout;
 
-   if (!(usage & GBM_BO_USE_CURSOR_64X64))
-      return NULL;
-   if (format != GBM_FORMAT_ARGB8888)
+   is_cursor = (usage & GBM_BO_USE_CURSOR) != 0 &&
+      format == GBM_FORMAT_ARGB8888;
+   is_scanout = (usage & GBM_BO_USE_SCANOUT) != 0 &&
+      format == GBM_FORMAT_XRGB8888;
+   if (!is_cursor && !is_scanout) {
+      errno = EINVAL;
       return NULL;
+   }
 
    bo = calloc(1, sizeof *bo);
    if (bo == NULL)
@@ -489,16 +791,7 @@ create_dumb(struct gbm_device *gbm,
    bo->handle = create_arg.handle;
    bo->size = create_arg.size;
 
-   memset(&map_arg, 0, sizeof(map_arg));
-   map_arg.handle = bo->handle;
-
-   ret = drmIoctl(dri->base.base.fd, DRM_IOCTL_MODE_MAP_DUMB, &map_arg);
-   if (ret)
-      goto destroy_dumb;
-
-   bo->map = mmap(0, bo->size, PROT_WRITE,
-                  MAP_SHARED, dri->base.base.fd, map_arg.offset);
-   if (bo->map == MAP_FAILED)
+   if (gbm_dri_bo_map(bo) == NULL)
       goto destroy_dumb;
 
    return &bo->base.base;
@@ -523,7 +816,7 @@ gbm_dri_bo_create(struct gbm_device *gbm,
    int dri_format;
    unsigned dri_use = 0;
 
-   if (usage & GBM_BO_USE_WRITE)
+   if (usage & GBM_BO_USE_WRITE || dri->image == NULL)
       return create_dumb(gbm, width, height, format, usage);
 
    bo = calloc(1, sizeof *bo);
@@ -550,13 +843,20 @@ gbm_dri_bo_create(struct gbm_device *gbm,
    case GBM_FORMAT_ABGR8888:
       dri_format = __DRI_IMAGE_FORMAT_ABGR8888;
       break;
+   case GBM_FORMAT_ARGB2101010:
+      dri_format = __DRI_IMAGE_FORMAT_ARGB2101010;
+      break;
+   case GBM_FORMAT_XRGB2101010:
+      dri_format = __DRI_IMAGE_FORMAT_XRGB2101010;
+      break;
    default:
-      return NULL;
+      errno = EINVAL;
+      goto failed;
    }
 
    if (usage & GBM_BO_USE_SCANOUT)
       dri_use |= __DRI_IMAGE_USE_SCANOUT;
-   if (usage & GBM_BO_USE_CURSOR_64X64)
+   if (usage & GBM_BO_USE_CURSOR)
       dri_use |= __DRI_IMAGE_USE_CURSOR;
 
    /* Gallium drivers requires shared in order to get the handle/stride */
@@ -568,7 +868,7 @@ gbm_dri_bo_create(struct gbm_device *gbm,
                               dri_format, dri_use,
                               bo);
    if (bo->image == NULL)
-      return NULL;
+      goto failed;
 
    dri->image->queryImage(bo->image, __DRI_IMAGE_ATTRIB_HANDLE,
                           &bo->base.base.handle.s32);
@@ -576,6 +876,10 @@ gbm_dri_bo_create(struct gbm_device *gbm,
                           (int *) &bo->base.base.stride);
 
    return &bo->base.base;
+
+failed:
+   free(bo);
+   return NULL;
 }
 
 static struct gbm_surface *
@@ -623,15 +927,18 @@ static struct gbm_device *
 dri_device_create(int fd)
 {
    struct gbm_dri_device *dri;
-   int ret;
+   int ret, force_sw;
 
    dri = calloc(1, sizeof *dri);
+   if (!dri)
+      return NULL;
 
    dri->base.base.fd = fd;
    dri->base.base.bo_create = gbm_dri_bo_create;
    dri->base.base.bo_import = gbm_dri_bo_import;
    dri->base.base.is_format_supported = gbm_dri_is_format_supported;
    dri->base.base.bo_write = gbm_dri_bo_write;
+   dri->base.base.bo_get_fd = gbm_dri_bo_get_fd;
    dri->base.base.bo_destroy = gbm_dri_bo_destroy;
    dri->base.base.destroy = dri_destroy;
    dri->base.base.surface_create = gbm_dri_surface_create;
@@ -640,7 +947,15 @@ dri_device_create(int fd)
    dri->base.type = GBM_DRM_DRIVER_TYPE_DRI;
    dri->base.base.name = "drm";
 
-   ret = dri_screen_create(dri);
+   force_sw = getenv("GBM_ALWAYS_SOFTWARE") != NULL;
+   if (!force_sw) {
+      ret = dri_screen_create(dri);
+      if (ret)
+         ret = dri_screen_create_sw(dri);
+   } else {
+      ret = dri_screen_create_sw(dri);
+   }
+
    if (ret)
       goto err_dri;
 
